from muteria.drivers.testgeneration.tools_by_languages.c.shadow_se.shadow_se \
                                                import TestcasesToolShadowSE 
StaticTestcaseTool = None
DynamicTestcaseTool = TestcasesToolShadowSE