
{% extends 'dashboard_base.html' %}

{% block title %}Home{% endblock %}
{% block header %}{% endblock %}

{% block content %}
  <div class="container">
     <div class="row">
      <div class="col-lg-1"></div>
      <div class="col-lg-10">
        <h1 style="text-align:center;"> Hello World</h1>
      </div>
     <div class="col-lg-1"></div>
  </div>
</div>
{% endblock %}