<html>
   <body>
      <form action = "http://localhost:5000/saveconf" method = "POST" 
            enctype = "multipart/form-data">
         <p>Name <input type = "text" name = "Name" /></p>
         <p>Physics <input type = "text" name = "Physics" /></p>
         <p>Chemistry <input type = "text" name = "chemistry" /></p>
         <p>Maths <input type ="text" name = "Mathematics" /></p>
         <p><input type = "file" name = "file" /></p>

         <p><input type = "submit" value = "submit" /></p>
      </form>
   </body>
</html>