from muteria.drivers.testgeneration.tools_by_languages.c.semu.semu \
                                                    import TestcasesToolSemu 
StaticTestcaseTool = None 
DynamicTestcaseTool = TestcasesToolSemu